/*

Copyright (C) 2002 J.D. Cole

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
  $Header: /home/jd/hpo/cvs-root/code/dld-projects/mpi/mpi_recv.cc,v 1.7 2002/12/11 21:00:09 jd Exp $
*/

#include <octave_mpi.h>

#include <mpi.h>
#include <octave/oct.h>
#include <octave/byte-swap.h>

#include <sstream>

#include <stdio.h>

static std::string
mpi_read_binary_data (std::istream& is, bool swap,
		      oct_mach_info::float_format fmt,
		      const std::string& filename, bool& global,
		      octave_value& tc, std::string& doc);

DEFUN_DLD(mpi_recv, args, , "MPI_Recv")
{
 if(!octave_mpi_verify_session())
    {
      error("mpi_recv: mpi has not been called or"
	    " has concluded the MPI session");

      return octave_value();
    }

  if (args.length() < 1)
    {
      error("mpi_recv: requires 1 argument.");
      
      return octave_value();
    }

  // Make sure a valid, that means that it is a real vector of some size
  // and that each of the values is less than the MPI_COMM_WORD size
  int sender_rank;

  // Should verify the rank is valid
  if (args(0).is_real_scalar())
    sender_rank = args(0).int_value();
  else
    {
      error("mpi_recv: bad sender");

      return octave_value();
    }

  int buffer_length;
  MPI_Status mpi_status;

  // Get the sent data length
  MPI_Recv((void*)&buffer_length, 1, MPI_INT,
	   sender_rank, 
	   OCTAVE_MPI_TAG_SETUP,
	   MPI_COMM_WORLD,
	   &mpi_status);


  // Now get the binary description of the data
  char buffer_data[buffer_length];

  MPI_Recv((void*)buffer_data, buffer_length, MPI_BYTE,
	   sender_rank, 
	   OCTAVE_MPI_TAG_DATA, 
	   MPI_COMM_WORLD,
	   &mpi_status);

  // Pull the entry _start and _length data off the end of the buffer
  std::basic_stringstream<char> recv_buffer;
  std::basic_streambuf<char> *rdbuf;

  rdbuf = recv_buffer.rdbuf();
  rdbuf->pubsetbuf(buffer_data, buffer_length);

  // Read the float format
  oct_mach_info::float_format flt_fmt;

  recv_buffer.read(X_CAST(char *,&flt_fmt),
		   sizeof(oct_mach_info::float_format));

  // Convert binary buffer_data back to an octave value
  octave_value_list retr_list;
  bool global;
  octave_value recv_value;
  std::string help;
  std::string var_name;

  while(!recv_buffer.eof())
    {
      var_name = mpi_read_binary_data (recv_buffer, false, 
				       flt_fmt,
				       "", global, recv_value, help);
      if(var_name.empty())
	break;
      else
	retr_list.append(recv_value);
	 
    }

  return retr_list;
}

/* 
----------------------------------------------------------------------------
   The code below this statement (specifically functions "get_save_type" and 
   "save_binary_data" from the file load-save.cc) has been taken from the 
   GNU/Octave source code and is covered by the following license. Please 
   honor it apropriately.

   Also thanks to Andy Jacobson for his work initially on octave-mpi which uses
   the Octave load/save commands to convert variables to binary, on which, this 
   effort is based. (See http://corto.icg.to.infn.it/andy/octave-mpi/)
----------------------------------------------------------------------------

Copyright (C) 1996, 1997 John W. Eaton

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

static std::string
mpi_read_binary_data (std::istream& is, bool swap,
		      oct_mach_info::float_format fmt,
		      const std::string& filename, bool& global,
		      octave_value& tc, std::string& doc)
{
  std::string retval;

  char tmp = 0;

  FOUR_BYTE_INT name_len = 0;
  FOUR_BYTE_INT doc_len = 0;

  doc.resize (0);

  // We expect to fail here, at the beginning of a record, so not
  // being able to read another name should not result in an error.

  is.read (X_CAST (char *, &name_len), 4);
  if (! is)
    return retval;
  if (swap)
    swap_4_bytes (X_CAST (char *, &name_len));

  char name[name_len+1];
  name[name_len] = '\0';
  if (! is.read (X_CAST (char *, name), name_len))
    goto data_read_error;
  retval = name;

  is.read (X_CAST (char *, &doc_len), 4);
  if (! is)
    goto data_read_error;
  if (swap)
    swap_4_bytes (X_CAST (char *, &doc_len));

  char tdoc[doc_len+1];
  tdoc[doc_len] = '\0';
  if (! is.read (X_CAST (char *, tdoc), doc_len))
    goto data_read_error;
  doc = tdoc;

  if (! is.read (X_CAST (char *, &tmp), 1))
    goto data_read_error;
  global = tmp ? 1 : 0;

  tmp = 0;
  if (! is.read (X_CAST (char *, &tmp), 1))
    goto data_read_error;

  switch (tmp)
    {
    case 1:
      {
	if (! is.read (X_CAST (char *, &tmp), 1))
	  goto data_read_error;
	double dtmp;
	read_doubles (is, &dtmp, X_CAST (save_type, tmp), 1, swap, fmt);
	if (error_state || ! is)
	  goto data_read_error;
	tc = dtmp;
      }
      break;

    case 2:
      {
	FOUR_BYTE_INT nr, nc;
	if (! is.read (X_CAST (char *, &nr), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &nr));
	if (! is.read (X_CAST (char *, &nc), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &nc));
	if (! is.read (X_CAST (char *, &tmp), 1))
	  goto data_read_error;
	Matrix m (nr, nc);
	double *re = m.fortran_vec ();
	int len = nr * nc;
	read_doubles (is, re, X_CAST (save_type, tmp), len, swap, fmt);
	if (error_state || ! is)
	  goto data_read_error;
	tc = m;
      }
      break;

    case 3:
      {
	if (! is.read (X_CAST (char *, &tmp), 1))
	  goto data_read_error;
	Complex ctmp;
	read_doubles (is, X_CAST (double *, &ctmp),
		      X_CAST (save_type, tmp), 2, swap, fmt);
	if (error_state || ! is)
	  goto data_read_error;
	tc = ctmp;
      }
      break;

    case 4:
      {
	FOUR_BYTE_INT nr, nc;
	if (! is.read (X_CAST (char *, &nr), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &nr));
	if (! is.read (X_CAST (char *, &nc), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &nc));
	if (! is.read (X_CAST (char *, &tmp), 1))
	  goto data_read_error;
	ComplexMatrix m (nr, nc);
	Complex *im = m.fortran_vec ();
	int len = nr * nc;
	read_doubles (is, X_CAST (double *, im),
		      X_CAST (save_type, tmp), 2*len, swap, fmt);
	if (error_state || ! is)
	  goto data_read_error;
	tc = m;
      }
      break;

    case 5:
      {
	FOUR_BYTE_INT len;
	if (! is.read (X_CAST (char *, &len), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &len));
	char s[len+1];
	if (! is.read (X_CAST (char *, s), len))
	  goto data_read_error;
	s[len] = '\0';
	tc = s;
      }
      break;

    case 6:
      {
	if (! is.read (X_CAST (char *, &tmp), 1))
	  goto data_read_error;
	double bas, lim, inc;
	if (! is.read (X_CAST (char *, &bas), 8))
	  goto data_read_error;
	if (swap)
	  swap_8_bytes (X_CAST (char *, &bas));
	if (! is.read (X_CAST (char *, &lim), 8))
	  goto data_read_error;
	if (swap)
	  swap_8_bytes (X_CAST (char *, &lim));
	if (! is.read (X_CAST (char *, &inc), 8))
	  goto data_read_error;
	if (swap)
	  swap_8_bytes (X_CAST (char *, &inc));
	Range r (bas, lim, inc);
	tc = r;
      }
      break;

    case 7:
      {
	FOUR_BYTE_INT elements;
	if (! is.read (X_CAST (char *, &elements), 4))
	  goto data_read_error;
	if (swap)
	  swap_4_bytes (X_CAST (char *, &elements));
	charMatrix chm (elements, 0);
	int max_len = 0;
	for (int i = 0; i < elements; i++)
	  {
	    FOUR_BYTE_INT len;
	    if (! is.read (X_CAST (char *, &len), 4))
	      goto data_read_error;
	    if (swap)
	      swap_4_bytes (X_CAST (char *, &len));
	    char tmp[len+1];
	    if (! is.read (X_CAST (char *, tmp), len))
	      goto data_read_error;
	    if (len > max_len)
	      {
		max_len = len;
		chm.resize (elements, max_len, 0);
	      }
	    tmp [len] = '\0';
	    chm.insert (tmp, i, 0);
	  }
	tc = octave_value (chm, true);
      }
      break;

    default:
    data_read_error:
      error ("load: trouble reading binary file `%s'", filename.c_str ());
      break;
    }

  return retval;
}
